package jdk.management.resource.internal.inst;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.nio.ByteBuffer;
import jdk.internal.instrumentation.InstrumentationMethod;
import jdk.management.resource.ResourceRequest;
import java.io.IOException;
import java.util.concurrent.CompletableFuture;
import jdk.management.resource.ResourceRequestDeniedException;
import jdk.management.resource.internal.ApproverGroup;
import jdk.management.resource.internal.ResourceIdImpl;
import jdk.management.resource.internal.FutureWrapper;
import jdk.management.resource.internal.CompletionHandlerWrapper;
import java.util.concurrent.Future;
import java.nio.channels.CompletionHandler;
import java.net.SocketAddress;
import java.net.InetSocketAddress;
import jdk.internal.instrumentation.InstrumentationTarget;

@InstrumentationTarget("sun.nio.ch.UnixAsynchronousSocketChannelImpl")
public class UnixAsynchronousSocketChannelImplRMHooks
{
    protected volatile InetSocketAddress localAddress;
    
    public UnixAsynchronousSocketChannelImplRMHooks() {
        this.localAddress = null;
    }
    
    @InstrumentationMethod
     <A> Future<Void> implConnect(SocketAddress socketAddress, A a, CompletionHandler<Void, ? super A> completionHandlerWrapper) {
        int n = this.localAddress != null;
        if (completionHandlerWrapper != null && n == 0) {
            completionHandlerWrapper = new CompletionHandlerWrapper(completionHandlerWrapper, this);
        }
        Future<Void> implConnect = this.implConnect(socketAddress, a, completionHandlerWrapper);
        if (implConnect != null && n == 0) {
            if (implConnect.isDone()) {
                ResourceIdImpl of = ResourceIdImpl.of(this.localAddress);
                ResourceRequest approver = ApproverGroup.SOCKET_OPEN_GROUP.getApprover(this);
                long request = 0L;
                Object ex = null;
                try {
                    request = approver.request(1L, of);
                    if (request < 1L) {
                        ex = new ResourceRequestDeniedException("Resource limited: too many open sockets");
                    }
                }
                catch (ResourceRequestDeniedException ex2) {
                    ex = ex2;
                }
                if (ex != null) {
                    approver.request(-request, of);
                    CompletableFuture o = new CompletableFuture();
                    o.completeExceptionally(ex);
                    implConnect = o;
                    try {
                        this.implClose();
                    }
                    catch (IOException ex3) {}
                }
                else {
                    approver.request(-(request - 1L), of);
                }
            }
            else {
                implConnect = new FutureWrapper(implConnect, this);
            }
        }
        return implConnect;
    }
    
    @InstrumentationMethod
     <V extends Number, A> Future<V> implRead(boolean b, ByteBuffer byteBuffer, ByteBuffer[] array, long n, TimeUnit timeUnit, A a, CompletionHandler<V, ? super A> completionHandlerWrapper) {
        ResourceIdImpl of = ResourceIdImpl.of(this.localAddress);
        ResourceRequest approver = ApproverGroup.SOCKET_READ_GROUP.getApprover(this);
        int remaining;
        if (b) {
            remaining = 0;
            for (int length = array.length, i = 0; i < length; ++i) {
                remaining += array[i].remaining();
            }
        }
        else {
            remaining = byteBuffer.remaining();
        }
        long max;
        try {
            max = Math.max(approver.request((long)remaining, of), 0L);
            if (max < (long)remaining) {
                throw new ResourceRequestDeniedException("Resource limited: insufficient bytes approved");
            }
        }
        catch (ResourceRequestDeniedException ex) {
            if (completionHandlerWrapper != null) {
                completionHandlerWrapper.failed(ex, a);
                return null;
            }
            CompletableFuture o = new CompletableFuture();
            o.completeExceptionally(ex);
            return o;
        }
        if (completionHandlerWrapper != null) {
            completionHandlerWrapper = new CompletionHandlerWrapper(completionHandlerWrapper, of, approver, max);
        }
        Future<Number> implRead = this.implRead(b, byteBuffer, array, n, timeUnit, a, completionHandlerWrapper);
        if (completionHandlerWrapper == null) {
            if (implRead.isDone()) {
                int intValue = 0;
                try {
                    intValue = ((Number)implRead.get()).intValue();
                }
                catch (InterruptedException | ExecutionException ex2) {}
                approver.request(-(max - (long)Math.max(0, intValue)), of);
            }
            else {
                implRead = new FutureWrapper(implRead, of, approver, max);
            }
        }
        return implRead;
    }
    
    @InstrumentationMethod
     <V extends Number, A> Future<V> implWrite(boolean b, ByteBuffer byteBuffer, ByteBuffer[] array, long n, TimeUnit timeUnit, A a, CompletionHandler<V, ? super A> completionHandlerWrapper) {
        ResourceIdImpl of = ResourceIdImpl.of(this.localAddress);
        ResourceRequest approver = ApproverGroup.SOCKET_WRITE_GROUP.getApprover(this);
        int remaining;
        if (b) {
            remaining = 0;
            for (int length = array.length, i = 0; i < length; ++i) {
                remaining += array[i].remaining();
            }
        }
        else {
            remaining = byteBuffer.remaining();
        }
        long max;
        try {
            max = Math.max(approver.request((long)remaining, of), 0L);
            if (max < (long)remaining) {
                throw new ResourceRequestDeniedException("Resource limited: insufficient bytes approved");
            }
        }
        catch (ResourceRequestDeniedException ex) {
            if (completionHandlerWrapper != null) {
                completionHandlerWrapper.failed(ex, a);
                return null;
            }
            CompletableFuture o = new CompletableFuture();
            o.completeExceptionally(ex);
            return o;
        }
        if (completionHandlerWrapper != null) {
            completionHandlerWrapper = new CompletionHandlerWrapper(completionHandlerWrapper, of, approver, max);
        }
        Future<Number> implWrite = this.implWrite(b, byteBuffer, array, n, timeUnit, a, completionHandlerWrapper);
        if (completionHandlerWrapper == null) {
            if (implWrite.isDone()) {
                int intValue = 0;
                try {
                    intValue = ((Number)implWrite.get()).intValue();
                }
                catch (InterruptedException | ExecutionException ex2) {}
                approver.request(-(max - (long)Math.max(0, intValue)), of);
            }
            else {
                implWrite = new FutureWrapper(implWrite, of, approver, max);
            }
        }
        return implWrite;
    }
    
    @InstrumentationMethod
    void implClose() throws IOException {
        try {
            this.implClose();
        }
        finally {
            if (this.localAddress != null) {
                ApproverGroup.SOCKET_OPEN_GROUP.getApprover(this).request(-1L, ResourceIdImpl.of(this.localAddress));
            }
        }
    }
}
